# BioPandas
# Authors: Sebastian Raschka <mail@sebastianraschka.com>
# Authors: Arian Jamasb <arian@jamasb.io>
# License: BSD 3 clause
# Project Website: http://rasbt.github.io/biopandas/
# Code Repository: https://github.com/rasbt/biopandas

"""
BioPandas module for working with mmCIF
files in pandas DataFrames.
"""

from .pandas_mmcif import PandasMmcif

__all__ = ["PandasMmcif"]
